
Analogs <- function(exp, obs, nanalogs = 3, fun_analog = "mean", data_dim = "data", lat_dim = "lat", lon_dim = "lon",
                    sdate_dim = "sdate", time_dim = "time", ncores = 1) {

  require(multiApply)
  require(CSTools)

  #-----------------------------------
  # Checkings
  #-----------------------------------
  # input exp and obs must be s2dv_cube objects
  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }

  # input exp and obs must be s2dv_cube objects
  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  if (!inherits(nanalogs, 'numeric')) {
    stop("Parameter 'nanalogs' must be of the class 'character'")
  }

  if (!inherits(fun_analog, 'character')) {
    stop("Parameter 'fun_analog' must be of the class 'character'")
  }

  # Do not allow synonims for lat (latitude), lon (longitude) and time (sdate) dimension names 
  if (is.na(match(lon_dim, names(dim(exp[[data_dim]])))) | is.na(match(lon_dim, names(dim(obs[[data_dim]]))))) {
    stop("Missing longitude dimension in 'exp' and/or 'obs', or does not match the parameter
         'lon_dim'")
  } 

  if (is.na(match(lat_dim, names(dim(exp[[data_dim]])))) | is.na(match(lat_dim, names(dim(obs[[data_dim]]))))) {
    stop("Missing latitude dimension in 'exp' and/or 'obs', or does not match the parameter
         'lat_dim'")
  }

  if (is.na(match(sdate_dim, names(dim(exp[[data_dim]])))) | is.na(match(sdate_dim, names(dim(obs[[data_dim]]))))) {
    stop("Missing start date dimension in 'exp' and/or 'obs', or does not match the parameter
         'sdate_dim'")
  }

  if (is.na(match(time_dim, names(dim(exp[[data_dim]])))) | is.na(match(time_dim, names(dim(obs[[data_dim]]))))) {
    stop("Missing time dimension in 'exp' and/or 'obs', or does not match the parameter
         'time_dim'")
  }

  # Ensure we have enough data to interpolate from high-res to coarse grid
  if ((attr(obs[[lat_dim]],"first_lat") > attr(exp[[lat_dim]],"first_lat")) | 
      (attr(obs[[lat_dim]],"last_lat") < attr(exp[[lat_dim]],"last_lat")) | 
      (attr(obs[[lon_dim]],"first_lon") > attr(exp[[lon_dim]],"first_lon")) |
      (attr(obs[[lon_dim]],"last_lon") < attr(exp[[lon_dim]],"last_lon"))) {
    stop("There are not enough data in 'obs'. Please to add more latitudes or ",
         "longitudes.")
  }

  # Select a function to apply to the analogs selected for a given observation
  stopifnot(fun_analog %in% c("mean", "wmean", "max", "min", "median"))

  #-----------------------------------
  # Interpolate high-res observations to the coarse grid
  #-----------------------------------
  grid_exp <- exp$source_files[1]  
  
  lonmin <- attr(exp[[lon_dim]],"first_lon")
  lonmax <- attr(exp[[lon_dim]],"last_lon")
  latmin <- attr(exp[[lat_dim]],"first_lat")
  latmax <- attr(exp[[lat_dim]],"last_lat")
  obs_interpolated <- Interpolation(exp = obs, target_grid = grid_exp, method_remap = "conservative",
                                    remap_region = c(lonmin, lonmax, latmin, latmax))
  lats <- obs[[lat_dim]]
  lons <- obs[[lon_dim]]

  #-----------------------------------
  # Reshape train and test
  #-----------------------------------
  res <- Apply(list(obs_interpolated$data, exp[[data_dim]], obs[[data_dim]]), 
               target_dims = list(c(time_dim, sdate_dim, lat_dim, lon_dim), 
               c(time_dim, sdate_dim, lat_dim, lon_dim), c(time_data, sdate_dim, lat_dim, lon_dim)), 
               fun = function(tr, te, ob) .analogs(train = tr, test = te, obs_hres = ob, k = nanalogs, 
               fun = fun_analog))$output1

  res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = lons, lat = lats))
  
  return(res_s2dv)
}


# For each element in test return the corresponding k-NN observations, or its mean if indicated in fun
.analogs <- function(train, test, obs_hres, k, fun = NULL) {

  require(FNN)
  require(multiApply)

  # Get knn indices for each test element
  knn.ind <- get_knn(train = train, test = test, k = k)  

  dist <- knn.ind$nn.dist
  idx <- knn.ind$nn.index
  names(dim(idx)) <- c(sdate_dim, "analog"); names(dim(dist)) <- c(sdate_dim, "analog")

  # Obtain the obs for the knn indices
  analogs <- Apply(idx, margins = "analog", fun = function(an) { obs_hres[an, , ] })$output1

  # Apply function for analogs
  if (!is.null(fun)) {
  
    margins_analogs <- names(dim(analogs))[names(dim(analogs)) != "analog"]
   
    if (fun == "wmean") {
      w <- 1 / dist 
      analogs <- Apply(list(analogs, dist), margins = list(margins_analogs, sdate_dim), 
                       fun = function(x, w) weighted.mean(x, w))$output1
    } else {   
      analogs <- Apply(analogs, margins = margins_analogs, fun = fun)$output1
    }
  }

  return(analogs)
}

# For each element in test, find the indices of the k nearest neigbhors in train
get_knn <- function(train, test, k) {
  # Reformat train and test as an array with (time, points)
  train <- aperm(Apply(train, target_dims = list(c(lat_dim, lon_dim)), output_dims = 'space', fun = function(x) { as.vector(x) })$output1)
  test <- aperm(Apply(test, target_dims = list(c(lat_dim, lon_dim)), output_dims = 'space', fun = function(x) { as.vector(x) })$output1)

  # For each row in test, find the k closest rows in train
  nsdates <- dim(train)[names(dim(train)) == sdate_dim]
  knn.ind <- lapply(1:nsdates, function(j) {
    te <- test[ j, , , drop = FALSE]
    tr <- train[ -j, , ]
    tr <- Apply(tr, target_dims = list(c(sdate_dim, time_dim)), output_dims = 't', fun = function(x) { as.vector(x) })$output1
    te <- Apply(te, target_dims = list(c(sdate_dim, time_dim)), output_dims = 't', fun = function(x) { as.vector(x) })$output1
    get.knnx(tr, te, k)
  })

  return(knn.ind)
}


